
#ifndef CPP_START_SYNC_QUEUE_H
#define CPP_START_SYNC_QUEUE_H

#include <mutex>
#include <list>
#include <thread>
#include <condition_variable>

template<typename T>
class SyncQueue {

private:
    int mMaxSize; // 同步队列最大数量
    std::mutex mMutex; // 互斥锁
    std::list<T> mQueue; // 缓冲区
    std::condition_variable mNotFull; // 没满的条件
    std::condition_variable mNotEmpty; // 不为空的条件
    bool mStop; // 停止标识

public:
    explicit SyncQueue(int maxSize)
            : mMaxSize(maxSize) {
    }

    /**
     * 入队
     * @param item 队列元素
     */
    void enqueue(const T &item) {
        add(item);
    }

    void enqueue(T &&item) {
        add(std::forward<T>(item));
    }


    template<typename F>
    void add(F &&f) {
        std::unique_lock<std::mutex> locker(mMutex);
        mNotFull.wait(locker, [this] { return mStop || mQueue.size() != mMaxSize; });
        if (mStop) return;
        mQueue.push_back(std::forward<F>(f));
        mNotEmpty.notify_all();
        mNotEmpty.notify_all();
    }


    /**
     * 出队，获取队列内的首个元素
     * @param item 元素
     */
    void dequeue(T &item) {
        std::unique_lock<std::mutex> locker(mMutex);
        mNotEmpty.wait(locker, [this] { return mStop || !mQueue.empty(); });
        if (mStop) return;
        item = mQueue.front();
        mQueue.pop_front();
        mNotFull.notify_all();
    }

    /**
     * 出队，获取队列内的首个元素
     * @param item 元素
     */
    void dequeue(std::list<T> &list) {
        std::unique_lock<std::mutex> locker(mMutex);
        mNotEmpty.wait(locker, [this] { return mStop || !mQueue.empty(); });
        if (mStop) return;
        list = std::move(mQueue);
        mNotFull.notify_all();
    }

    /**
     * 当前队列数量
     * @return  当前队列数量
     */
    size_t size() {
        std::lock_guard<std::mutex> locker(mMutex);
        return mQueue.size();
    }

    /**
     * 查看最大容量
     * @return  最大容量
     */
    int getMaxSize() const {
        return mMaxSize;
    }

    void stop() {
        {
            std::lock_guard<std::mutex> locker(mMutex);
            mStop = true;
        }
        mNotFull.notify_all();
        mNotEmpty.notify_all();
    }

};


#endif //CPP_START_SYNC_QUEUE_H
