#include <iostream>
#include <thread>

std::mutex g_lock;

void func() {
    std::lock_guard<std::mutex> locker(g_lock);
    std::this_thread::sleep_for(std::chrono::seconds(2));
}

void thread_test() {
    auto start = std::chrono::duration_cast<std::chrono::milliseconds>(
            std::chrono::system_clock::now().time_since_epoch()).count();
    std::cout << "start time: " << start / 1000 << std::endl;
    std::thread t0(func);
    std::thread t1(func);
    std::thread t2(func);
    t0.join(); // 阻塞执行
    t1.detach(); // 分离线程何对象，之后就无法何线程发送联系
    t2.join();
    auto end = std::chrono::duration_cast<std::chrono::milliseconds>(
            std::chrono::system_clock::now().time_since_epoch()).count();
    std::cout << "  end time: " << end / 1000 << std::endl;
}

#include "base/sync_queue.h"

void sync_test() {
    // 同步队列
    auto queue = new SyncQueue<int>(3);

    // 每2秒取出进行消费
    std::thread producer([&]() {
        int i = 0;
        while (i < 9) {
            queue->dequeue(i);
            std::this_thread::sleep_for(std::chrono::seconds(2));
            std::cout << "dequeue: " << i << std::endl;
        }
    });

    // 持续写入
    std::thread consumer([&]() {
        for (int i = 0; i < 10; ++i) {
            queue->enqueue(i);
            std::cout << "enqueue: " << i << std::endl;
        }
    });

    producer.join();
    consumer.join();
}

#include <chrono>
#include <iomanip>
#include "util/stop_watch.h"

void chrono_test() {

    StopWatch stopWatch;

    std::chrono::minutes t1(10);
    std::chrono::seconds t2(60);
    std::chrono::seconds t3 = t1 - t2;
    std::cout << "duration time: " << t3.count() << std::endl;

    std::cout << "\ncost time: " << stopWatch.elapsed<std::chrono::microseconds>() << " us\n" << std::endl;

    std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
    // 将 time_point 转为 ctime
    std::time_t last = std::chrono::system_clock::to_time_t(now - std::chrono::hours(24));
    std::time_t next = std::chrono::system_clock::to_time_t(now + std::chrono::hours(24));

    // 使用put_time格式化时间
    std::cout
            << "One day ago, the time was  "
            << std::put_time(std::localtime(&last), "%F %T")
            << "\n"
            << "NExt day ago, the time is  "
            << std::put_time(std::localtime(&next), "%F %T")

            << std::endl;

    std::cout << "\ncost time: " << stopWatch.elapsed() << " ms\n" << std::endl;

}

#include "base/thread_pool.h"

void threadpool_test() {
    ThreadPool pool(3, 10);

    std::thread t0([&pool] {
        for (int i = 0; i < 100; ++i) {
            auto thdId = std::this_thread::get_id();
            pool.add([thdId] {
                std::cout << "thread 1 id:" << thdId << std::endl;
            });
        }
    });

    std::thread t1([&pool] {
        for (int i = 0; i < 100; ++i) {
            auto thdId = std::this_thread::get_id();
            pool.add([thdId] {
                std::cout << "thread 2 id:" << thdId << std::endl;
            });
        }
    });

    std::this_thread::sleep_for(std::chrono::seconds(2));
//    getchar();

    pool.stop();
    t0.join();
    t1.join();
}

#include "aspect/aspect.h"
#include "test/time_elapsed_aspect.h"
#include "test/log_aspect.h"

void HT(int a) {
    std::cout << "real HT function " << a << std::endl;
}

void aop_test() {
    using namespace std;

    aop<LogAspect, TimeElapsedAspect>([](int a) {
        std::thread t([a]() {
            std::this_thread::sleep_for(std::chrono::seconds(1));
            std::cout << "run function, arg is " << a << std::endl;
        });
        t.join();
    }, 1);

    cout << endl;

    std::function<void(int)> f = std::bind(&HT, std::placeholders::_1);
    aop<LogAspect>(f, 100);
}


#include "base/any.h"

void any_test() {
    Any n;
    auto r = n.IsNull();//true
    string s1 = "hello";
    n = s1;
    n = "world";
    n.AnyCast<int>(); //can not cast int to string
    Any n1 = 1;
    n1.Is<int>(); //true
}

#include "ioc/container.h"

void ioc_test() {

    struct Base {
        virtual ~Base() = default;

        virtual void Func() = 0;
    };

    struct DerivedB : public Base {
        DerivedB(int a, double b) : m_a(a), m_b(b) {
        }

        ~DerivedB() override = default;

        void Func() override {
            cout << "DerivedB run :" << m_a + m_b << endl;
        }

    private:
        int m_a;
        double m_b;
    };

    struct DerivedC : public Base {
        ~DerivedC() override = default;

        void Func() override {
            std::cout << "DerivedC " << std::endl;
        };
    };

    struct A {
        A(Base *ptr) : m_ptr(ptr) {
        }

        ~A() {
            if (m_ptr != nullptr) {
                delete m_ptr;
                m_ptr = nullptr;
            }
        }

        void Func() {
            m_ptr->Func();
        }

    private:
        Base *m_ptr;
    };

    IocContainer ioc;

    // TODO 可通过CRTP奇怪的重复模板模式进行自动注册

    // 注册依赖关系
    ioc.RegisterType<A, DerivedC>("C");
    ioc.RegisterType<A, DerivedB, int, double>("B");

    // 通过容器进行初始化
    auto pc = ioc.ResolveShared<A>("C");
    // 注意DerivedB的参数int,double
    auto pb = ioc.ResolveShared<A>("B", 1, 2.0);

    pc->Func();
    pb->Func();
}

#include "bus/message_bus.h"

void message_test() {
    MessageBus bus;
    //注册消息
    bus.Attach([](int a) { cout << "no reference: " << a << endl; });
    bus.Attach([](int &a) { cout << "lvalue reference: " << a << endl; });
    bus.Attach([](int &&a) { cout << "rvalue reference: " << a << endl; });
    bus.Attach([](const int &a) { cout << "const lvalue reference: " << a << endl; });
    bus.Attach([](int a) {
        cout << "no reference has return value and key: " << a << endl;
        return a;
    }, "a");

    int i = 2;
    bus.SendReq<void, int>(2);
    bus.SendReq<int, int>(2, "a");
    bus.SendReq<void, int &>(i);
    bus.SendReq<void, const int &>(2);
    bus.SendReq<void, int &&>(2);

    bus.Remove<void, int>();
    bus.Remove<int, int>("a");
    bus.Remove<void, int &>();
    bus.Remove<void, const int &>();
    bus.Remove<void, int &&>();

    bus.SendReq<void, int>(2);
    bus.SendReq<int, int>(2, "a");
    bus.SendReq<void, int &>(i);
    bus.SendReq<void, const int &>(2);
    bus.SendReq<void, int &&>(2);

}

int main() {
//     线程测试
    thread_test();
//     同步队列测试
    sync_test();
//     时间工具测试
    chrono_test();
//     线程池测试
    threadpool_test();
//     AOP测试
    aop_test();
//     Any测试
    any_test();
//     IoC测试
    ioc_test();
//     总线测试
    message_test();
    return 0;
}